#!/usr/bin/env python
# -*- coding: utf-8 -*-

# A HUD that is easy to use.
# Copyright Gummbum, 4/18/2017. Free software, no warranty express or implied.

"""hudlight.py - a basic HUD with a lot of conveniences

hud = HUD()

hud.add('fps', '{} fps', 0)
hud.update_item('fps', int(clock.get_fps()))

hud.add('date', '{} {}', time, date)
hud.update_item('date', time, date)

hud.add('mouse', '{0[0]}, {0[1]}', (0, 0))
hud.update_item('mouse', pygame.mouse.get_pos())

hud.draw(pygame.display.get_surface())

See hudlight-demo1.py for a typical usage.

Despite the seeming simplicity of the above examples, one can do quite a lot with this HUD. See the internal demo in the
__main__ code block at the end of this module for a very elaborate example. But keep in mind that the builtin demo is
not typical of HUD implementations. It is very elaborate because the HUD is the GUI which presents all the HUD features.
HUDs are usually much simpler, displaying a username and score, etc. To play with the demo, run: python hudlight.py.


BUGS:
1. The gradient function uses pygame.PixelArray. PixelArray has a bug that prevents modification of surfaces with
SRCCOLORKEY, so gradient colors effect cannot be achieved unless Hud.ANTIALIAS is True.
"""

import collections

import pygame

from fonter import load_font, make_colors, render


__all__ = ['HUDNameExists', 'HUDNameNotFound', 'HUDBadArgs', 'HUD']


class HUDNameExists(BaseException):
    """HUD non-unique name in items list"""
    pass


class HUDNameNotFound(BaseException):
    """HUD name not found in items list"""
    pass


class HUDBadArgs(BaseException):
    """HUD update_item() arg count does not match args given to add()"""
    pass


class HUD(object):
    """HUD class for a basic Heads Up Display

    Not supported:
        - Per-item details (font, size, style, antialias, color); see Caveats for more robust usage.

    Class variables:
        The following can be set at the class level. Once the instance is created, use the set_*() methods
        instead.

        FONT: font (pygame font object; or file font or system font spec)
        SIZE: font point size (int)
        BOLD: bold style (True or False; only relevant for system fonts)
        ITALIC: italic style (True or False; only relevant for system fonts)
        ANTIALIAS: render antialiased fonts (True or False)
        FGCOLOR: font forground color (any valid pygame.Color)
        BGCOLOR: font background color (any valid pygame.Color; or None for transparent)
        GCOLOR: font gradient color (any valid pygame.Color; or None for no gradient)
        COLORKEY: surface colorkey (any valid pygame.Color; or None for opaque)
        ALPHA: surface alpha (0 to 255; or None to disable)
        SHADOWCOLOR: render a drop shadow (any valid pygame.Color; or None to disable)
        OUTLINECOLOR: render an outline (any valid pygame.Color; or None to disable)
        X: X position of the HUD (pixels as int)
        Y; Y position of the HUD (pixels as int)
        SPACE; pixels between entries (pixels as int)

        Changing the class or instance variables during runtime does not trigger rendering or font creation on the
        instance.
        - To trigger font creation and rendering on a FONT or SIZE change, call set_font() without args.
        - To trigger rendering, use the set_*() methods.
        - X, Y, and SPACE take effect immediately IFF they have not been set locally on the instance.

    Tips and tricks:
        Note the difference between set_fgcolor(None) and set_bgcolor(None). The latter has the side effect of
        making the background transparent.

        ALPHA (aka surface alpha) only has an effect when BGCOLOR is used (i.e. not None). This is due to the
        behavior of the Font, not HUD. If you want to use alpha and keep a transparent background, then
        set_bgcolor() and set_colorkey() to the same color; this creates a surface that is compatible with
        surface alpha.

        Combining bgcolor and colorkey do not always play well with antialias. Run the demo in __main__: select
        bgcolor=skyblue, colorkey=True, and then toggle antialias on and off to compare.

        The font handling in set_font() leans towards forgiveness. Unfortunately, this may also result in some
        fuzzy results (rare) and error masking (more common).

        set_bold() and set_italic() (BOLD and ITALIC styles) are only relevant for system fonts. This is
        intentional. pygame supports a simulated rendering for these styles, but it generally doesn't look good.
        Internally, the HUD recreates the system font (via set_font()) when bold or italic changes. The results
        generally look better than the method pygame uses to simulate them. If you want to use the simulated
        styles anyway (e.g. for a file font that doesn't provide a styled font) you can hud._font.set_bold(True)
        or hud._font.set_italic(True), and trigger re-rendering with hud.set_dirty = True.

        When a file font does not exist, set_font() falls back on SysFont. This can mask programmer errors in
        file font paths. Therefore, vet your paths at run time.

        It is possible that a font file might clash with a SysFont. To override the logic in set_font(), create
        the font with pygame.font.Font or pygame.font.SysFont and supply the object to set_font().

        If you want per-item details (font, size, style, antialias, color) or want to use HUD in a more robust
        dashboard-like way, you can make multiple HUDs with custom X, Y positions. If you want to calculate
        their size for positioning, use the get_size() method, keeping in mind that changing SysFont styles,
        font file, etc. will change the rendered size.

        The font items are rendered in the order they are added. Order can be rearranged at any time by
        modifying the hud.order list, which is simply a list of the item names.
    """
    FONT = 'sans'           # font file or SysFont name
    SIZE = 18               # font size
    BOLD = False            # bold style
    ITALIC = False          # italic style
    ANTIALIAS = True        # antialias rendering (if False, gradient will not work)
    FGCOLOR = 'gold'        # foreground color
    BGCOLOR = None          # background color
    GCOLOR = None           # gradient color
    COLORKEY = None         # source colorkey
    ALPHA = None            # surface alpha
    SHADOWCOLOR = None      # drop shadow color
    OUTLINECOLOR = None     # outline color
    X = 10                  # x position of the HUD
    Y = 10                  # y position of the HUD
    SPACE = 3               # pixels between HUD items

    def __init__(self, font=None, size=None, bold=None, italic=None, fgcolor=None, bgcolor=None, gcolor=None, ocolor=None, scolor=None):
        """create a basic HUD

        If any of the args are not specified the class variables are used.

        :param font: a pygame Font object, or the name of a font file or system font
        :param size: point size of the font
        :param bold: boolean, only relevant for a sysfont
        :param italic: boolean, only relevant for a sysfont
        :param fgcolor: text foreground color; default is HUD.FGCOLOR
        :param bgcolor: text background color; default is HUD.BGCOLOR (None means transparent)
        :param gcolor: text gradient color; default is HUD.GCOLOR
        :param ocolor: text outline color; default is HUD.OCOLOR
        :param scolor: text shadow color; default is HUD.SCOLOR
        :return: HUD
        """
        self.order = []     # list to order rendering
        self.items = {}     # HUD data
        self.dirty = False

        self._font = None
        self._font_info = None, None, None, None
        self._fgcolor = None
        self._bgcolor = None
        self._gcolor = None
        self._scolor = None
        self._ocolor = None
        self._colorkey = None

        self.set_fgcolor(fgcolor if fgcolor else self.FGCOLOR)
        self.set_bgcolor(bgcolor if bgcolor else self.BGCOLOR)
        self.set_gradient(gcolor if gcolor else self.GCOLOR)
        self.set_outline(ocolor if ocolor else self.OUTLINECOLOR)
        self.set_shadow(scolor if scolor else self.SHADOWCOLOR)
        self.set_font(font, size, bold, italic)

        self.perf = collections.deque()
        self.perf_history_secs = 1

    def fps(self):
        return len(self.perf) / self.perf_history_secs

    def get_size(self):
        """get the total rendered size of the HUD (may change depending on style, etc.)

        self.SPACES

        :return: w, h
        """
        total_w = 0
        total_h = 0
        for item in self.items.values():
            w, h = item[0].get_size()
            # inline is much faster than max()
            total_w = w if w > total_w else total_w
            total_h = h if h > total_h else total_h
            total_h += self.SPACE
        total_h -= self.SPACE
        return total_w, total_h

    def get_font(self):
        """get the current font object

        :return: pygame.font.Font
        """
        return self._font

    def set_font(self, font=None, size=None, bold=None, italic=None):
        """change font

        If any of the args are not specified the instance variables are used. The instance variables are
        updated to reflect the supplied values.

        Font creation and re-rendering of items is triggered.

        :param font: a pygame Font object, or the name of a font file or system font
        :param size: point size of the font
        :param bold: boolean, only relevant for a sysfont
        :param italic: boolean, only relevant for a sysfont
        :return: None
        """
        font = self.FONT if font is None else font
        size = self.SIZE if size is None else size
        bold = self.BOLD if bold is None else bold
        italic = self.ITALIC if italic is None else italic

        new_info = font, size, bold, italic
        if self._font_info != new_info:
            self._font_info = new_info
            self.FONT = font
            self.SIZE = size
            self.BOLD = bold
            self.ITALIC = italic
            if isinstance(font, pygame.font.Font):
                self._font = font
            else:
                self._font = load_font(font, size, bold, italic)
            self.dirty = True

    def set_font_name(self, name):
        """change the font source

        :param name: a pygame Font object, or the name of a font file or system font
        :return: None
        """
        if self.FONT != name:
            self.FONT = name
            if self._font is not None:
                self.set_font(font=name)

    def set_size(self, size):
        """change the size

        :param size: int
        :return: None
        """
        if self.SIZE != size:
            self.SIZE = size
            if self._font is not None:
                self.set_font(size=size)

    def set_bold(self, boolean):
        """change the bold style

        :param boolean: only relevant for a sysfont
        :return: None
        """
        if self.BOLD != boolean:
            self.BOLD = boolean
            if self._font is not None:
                self.set_font(bold=boolean)

    def set_italic(self, boolean):
        """change the italic style

        :param boolean: only relevant for a sysfont
        :return: None
        """
        if self.ITALIC != boolean:
            self.ITALIC = boolean
            if self._font is not None:
                self.set_font(italic=boolean)

    def set_fgcolor(self, color):
        """change foreground color

        All items will be re-rendered.

        :param color: a pygame.Color, color sequence, or color name
        :return: None
        """
        if color is None:
            return

        new_color = make_colors(color)[0]
        if self._fgcolor != new_color:
            self._fgcolor = new_color
            self.FGCOLOR = color
            self.dirty = True

    def set_bgcolor(self, color):
        """change background color

        All items will be re-rendered. If color is None, the default will be set to transparent.

        :param color: a pygame.Color, color sequence, or color name
        :return: None
        """
        new_color = make_colors(color)[0]
        if self._bgcolor != new_color:
            self._bgcolor = new_color
            self.BGCOLOR = color
            self.dirty = True

    def set_gradient(self, color=None):
        """change gradient color

        All items will be re-rendered. If color is None, the gradient effect will be removed.

        :param color: a pygame.Color, color sequence, or color name
        :return: None
        """
        new_color = make_colors(color)[0]
        if self._gcolor != new_color:
            self._gcolor = new_color
            self.dirty = True

    def set_shadow(self, color=None):
        """render a drop shadow

        All items will be re-rendered. If color is None, the shadow effect will be removed.

        :param color: a pygame.Color, color sequence, or color name; if None, it is disabled
        :return: None
        """
        new_color = make_colors(color)[0]
        if self._scolor != new_color:
            self._scolor = new_color
            self.dirty = True

    def set_outline(self, color=None):
        """render an outline

        All items will be re-rendered. If color is None, the outline effect will be removed.

        :param color: a pygame.Color, color sequence, or color name; if None, it is disabled
        :return: None
        """
        new_color = make_colors(color)[0]
        if self._ocolor != new_color:
            self._ocolor = new_color
            self.dirty = True

    def set_colorkey(self, color=None):
        """set surface colorkey

        All items will be re-rendered. If color is None the colorkey is removed.

        :param color: pygame.Color or None
        :return: None
        """
        new_color = make_colors(color)[0]
        if self._colorkey != new_color:
            self._colorkey = new_color
            self.COLORKEY = color
            self.dirty = True

    def set_alpha(self, alpha):
        """change the surface alpha value

        Only affective if a background color is used. See "Tips and tricks" in the class docstring.

        :param alpha: int, 0 to 255; if None it is disabled
        :return: None
        """
        if self.ALPHA != alpha:
            self.ALPHA = alpha
            for item in self.items.values():
                item[0].set_alpha(alpha)

    def set_antialias(self, boolean):
        """render antialiased font

        Re-rendering of items is triggered.

        :param boolean: True or False
        :return: None
        """
        if self.ANTIALIAS != boolean:
            self.ANTIALIAS = boolean
            self.dirty = True

    def add(self, name, text, *args):
        """add a HUD item

        HUD items are rendered in the order they are added. Order can be rearranged at any time by modifying
        hud.order, which is simply a list of the item names.

        :param name: unique ID
        :param text: format text compatible with str.format()
        :param args: arguments for str.format()
        :return: None
        """
        if name in self.items:
            raise HUDNameExists('HUD non-unique name {}'.format(name))
        item = [None, text, args]
        self._render(item)
        self.order.append(name)
        self.items[name] = item

    def update_item(self, name, *args):
        """update the value of an item

        :param name: item's name
        :param args: item's new value
        :return: None
        """
        try:
            item = self.items[name]
        except KeyError:
            raise HUDNameNotFound('HUD name not found {}'.format(name))
        if len(args) != len(item[2]):
            raise HUDBadArgs('{}: arg count {} != {}'.format(name, len(args), len(item[2])))
        # if args != item[2]:     # <<< NO, does not compare deeper items
        if [a for a, b in zip(args, item[2]) if a != b]:
            item[2] = args
            self._render(item)

    def draw(self, surf):
        """draw all the items

        :param surf: target surface
        :return: None
        """
        if self.dirty:
            for item in self.items.values():
                self._render(item)
            self.dirty = False

        items = self.items
        x = self.X
        y = self.Y
        for key in self.order:
            img = items[key][0]
            surf.blit(img, (x, y))
            y += img.get_height() + self.SPACE

    def _render(self, item):
        item[0] = render(item[1].format(*item[2]), self._font, self._fgcolor,
                                self._bgcolor, self._gcolor, self._scolor, self._ocolor,
                                self._colorkey, self.ALPHA, self.ANTIALIAS)
        t0 = pygame.time.get_ticks()
        t1 = t0 - self.perf_history_secs * 1000
        perf = self.perf
        popleft = perf.popleft
        self.perf.append(t0)
        while perf[0] < t1:
            popleft()
